from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Union
from xml.etree.ElementTree import ElementTree
from xml.etree.ElementTree import indent as indent_svg


@dataclass(frozen=True)
class _Control:
    pass


@dataclass(frozen=True)
class Quit(_Control):
    """
    Cause the application to quit.

    The quit occurs at the beginning of the next render cycle,
    so all other events that are due to be processed in the current cycle
    will be processed before the application exits.
    """


@dataclass(frozen=True)
class Bell(_Control):
    """
    Cause the terminal to emit a bell sound.

    The bell occurs at the beginning of the next render cycle,
    so all other events that are due to be processed in the current cycle
    will be processed before the sound is played.
    """


@dataclass(frozen=True)
class Screenshot(_Control):
    """
    Take a "screenshot" of the rendered UI,
    using the given `handler` callback function.
    The screenshot is passed to the `handler` as an
    [`ElementTree`][xml.etree.ElementTree.ElementTree]
    containing an SVG representation of the UI.

    The screenshot is taken at the beginning of the next render cycle,
    so all other events that are due to be processed in the current cycle
    will be processed before the screenshot is taken
    (but the screenshot will still be of the UI from *before* the next render occurs!).
    """

    handler: Callable[[ElementTree], None]

    @classmethod
    def to_file(cls, path: Path, indent: int | None = None) -> Screenshot:
        """
        A convenience method for producing a `Screenshot`
        that writes the resulting SVG to the given `path`.

        Parameters:
            path: The path to write the SVG to.
            indent: The number of spaces to indent the SVG by (for readability).
                If `None`, the SVG will not be indented.
        """

        def handler(et: ElementTree) -> None:
            if indent:
                indent_svg(et, space=" " * indent)
            with path.open("w") as f:
                et.write(f, encoding="unicode")

        return cls(handler=handler)


@dataclass(frozen=True)
class ToggleBorderHealing(_Control):
    """
    Toggle whether border healing occurs.
    """


AnyControl = Union[
    Quit,
    Bell,
    Screenshot,
    ToggleBorderHealing,
]



from __future__ import annotations

from collections.abc import Iterator

from more_itertools import take
from pydantic import Field, NonNegativeInt
from structlog import get_logger

from counterweight._utils import halve_integer, partition_int
from counterweight.cell_paint import wrap_cells
from counterweight.components import Component
from counterweight.elements import AnyElement
from counterweight.geometry import Position
from counterweight.styles.styles import BorderEdge
from counterweight.types import ForbidExtras

logger = get_logger()


class Rect(ForbidExtras):
    x: int = Field(default=0)
    y: int = Field(default=0)
    width: NonNegativeInt = Field(default=0)
    height: NonNegativeInt = Field(default=0)

    def expand_by(self, edge: Edge) -> Rect:
        return Rect(
            x=self.x - edge.left,
            y=self.y - edge.top,
            width=self.width + edge.left + edge.right,
            height=self.height + edge.top + edge.bottom,
        )

    def x_range(self) -> range:
        return range(self.x, self.x + self.width)

    def y_range(self) -> range:
        return range(self.y, self.y + self.height)

    @property
    def left(self) -> int:
        return self.x

    @property
    def right(self) -> int:
        return self.x + self.width - 1

    @property
    def top(self) -> int:
        return self.y

    @property
    def bottom(self) -> int:
        return self.y + self.height - 1

    def left_edge(self) -> list[Position]:
        left = self.left
        return [Position(left, y) for y in self.y_range()]

    def right_edge(self) -> list[Position]:
        right = self.right
        return [Position(right, y) for y in self.y_range()]

    def top_edge(self) -> list[Position]:
        top = self.top
        return [Position(x, top) for x in self.x_range()]

    def bottom_edge(self) -> list[Position]:
        bottom = self.bottom
        return [Position(x, bottom) for x in self.x_range()]

    def __contains__(self, item: object) -> bool:
        if isinstance(item, Position):
            return item.x in self.x_range() and item.y in self.y_range()
        else:
            return False


class Edge(ForbidExtras):
    left: int = Field(default=0)
    right: int = Field(default=0)
    top: int = Field(default=0)
    bottom: int = Field(default=0)


class BoxDimensions(ForbidExtras):
    content: Rect = Field(default_factory=Rect)
    margin: Edge = Field(default_factory=Edge)
    border: Edge = Field(default_factory=Edge)
    padding: Edge = Field(default_factory=Edge)

    def padding_border_margin_rects(self) -> tuple[Rect, Rect, Rect]:
        padding = self.content.expand_by(self.padding)
        border = padding.expand_by(self.border)
        margin = border.expand_by(self.margin)
        return padding, border, margin

    def left_edge_width(self) -> int:
        return self.margin.left + self.border.left + self.padding.left

    def right_edge_width(self) -> int:
        return self.margin.right + self.border.right + self.padding.right

    def top_edge_width(self) -> int:
        return self.margin.top + self.border.top + self.padding.top

    def bottom_edge_width(self) -> int:
        return self.margin.bottom + self.border.bottom + self.padding.bottom

    def horizontal_edge_width(self) -> int:
        return self.left_edge_width() + self.right_edge_width()

    def vertical_edge_width(self) -> int:
        return self.top_edge_width() + self.bottom_edge_width()

    def width(self) -> int:
        return self.content.width + self.horizontal_edge_width()

    def height(self) -> int:
        return self.content.height + self.vertical_edge_width()


class LayoutBox(ForbidExtras):
    element: AnyElement
    dims: BoxDimensions = Field(default_factory=BoxDimensions)
    parent: LayoutBox | None
    children: list[LayoutBox] = Field(default_factory=list)

    def walk_from_bottom(self) -> Iterator[LayoutBox]:
        for child in self.children:
            yield from child.walk_from_bottom()
        yield self

    def walk_elements_from_bottom(self) -> Iterator[AnyElement]:
        yield from (node.element for node in self.walk_from_bottom())

    def walk_levels(self) -> Iterator[LayoutBox]:
        yield self

        next_level: list[LayoutBox] = []
        for child in self.children:
            c = child.walk_levels()
            yield take(1, c)[0]
            next_level.extend(c)
        yield from next_level

    def compute_layout(self) -> None:
        top_down = list(self.walk_levels())

        for node in reversed(top_down):
            node.first_pass()

        for node in top_down:
            node.second_pass()

    def first_pass(self) -> None:
        style = self.element.style
        layout = style.layout

        # transfer margin/border/padding to dimensions
        self.dims.margin.top = style.margin.top
        self.dims.margin.bottom = style.margin.bottom
        self.dims.margin.left = style.margin.left
        self.dims.margin.right = style.margin.right

        self.dims.border.top = 1 if style.border and BorderEdge.Top in style.border.edges else 0
        self.dims.border.bottom = 1 if style.border and BorderEdge.Bottom in style.border.edges else 0
        self.dims.border.left = 1 if style.border and BorderEdge.Left in style.border.edges else 0
        self.dims.border.right = 1 if style.border and BorderEdge.Right in style.border.edges else 0

        self.dims.padding.top = style.padding.top
        self.dims.padding.bottom = style.padding.bottom
        self.dims.padding.left = style.padding.left
        self.dims.padding.right = style.padding.right

        # text boxes with auto width get their width from their content (no wrapping)
        # TODO: revisit this, kind of want to differentiate between "auto" and "flex" here, or maybe width=Weight(1) ?
        if self.element.type == "text" and self.element.style.typography.wrap == "none":
            if style.span.width == "auto":
                self.dims.content.width = max(
                    (
                        len(line)
                        for line in wrap_cells(
                            cells=self.element.cells,
                            wrap=style.typography.wrap,
                            width=100_000,  # any large number
                        )
                    ),
                    default=0,
                )
            if style.span.height == "auto":
                self.dims.content.height = len(
                    wrap_cells(
                        cells=self.element.cells,
                        wrap=style.typography.wrap,
                        width=100_000,  # any large number
                    )
                )

        if style.span.width != "auto":  # i.e., if it's a fixed width
            self.dims.content.width = style.span.width
        if style.span.height != "auto":  # i.e., if it's a fixed height
            self.dims.content.height = style.span.height

        num_gaps = max(sum(1 for child in self.children if child.element.style.layout.position == "relative") - 1, 0)

        # grow to fit children with fixed sizes
        if style.span.width == "auto":
            if layout.direction == "row":
                self.dims.content.width += num_gaps * layout.gap_children

            for child_box in self.children:
                child_element = child_box.element
                child_style = child_element.style
                child_layout = child_style.layout

                # if child_style.span.width != "auto" or (
                #     child_element.type == "text" and child_style.typography.wrap == "none"
                # ):
                if child_box.dims.content.width != 0:  # i.e., it has been set
                    if child_layout.position == "relative":
                        if layout.direction == "row":
                            # We are growing the box to the right
                            self.dims.content.width += child_box.dims.width()
                        elif layout.direction == "column":
                            # The box is as wide as its widest child
                            self.dims.content.width = max(self.dims.content.width, child_box.dims.width())

        if style.span.height == "auto":
            if layout.direction == "column":
                self.dims.content.height += num_gaps * layout.gap_children

            for child_box in self.children:
                child_element = child_box.element
                child_style = child_element.style
                child_layout = child_style.layout

                # if child_style.span.height != "auto" or (
                #     child_element.type == "text" and child_style.typography.wrap == "none"
                # ):
                if child_box.dims.content.height != 0:  # i.e., it has been set
                    if child_layout.position == "relative":
                        if layout.direction == "column":
                            # We are growing the box downward
                            self.dims.content.height += child_box.dims.height()
                        elif layout.direction == "row":
                            # The box is as tall as its tallest child
                            self.dims.content.height = max(self.dims.content.height, child_box.dims.height())

    def second_pass(self) -> None:
        style = self.element.style
        layout = style.layout
        parent = self.parent

        # TODO: positions

        # handle align self
        if layout.position == "relative" and parent:
            if parent.element.style.layout.direction == "row":
                match layout.align_self:
                    case "center":
                        self.dims.content.y += (parent.dims.content.height - self.dims.height()) // 2
                    case "end":
                        self.dims.content.y += parent.dims.content.height - self.dims.height()
                    case "stretch":
                        self.dims.content.height = parent.dims.content.height - self.dims.vertical_edge_width()
            elif parent.element.style.layout.direction == "column":
                match layout.align_self:
                    case "center":
                        self.dims.content.x += (parent.dims.content.width - self.dims.width()) // 2
                    case "end":
                        self.dims.content.x += parent.dims.content.width - self.dims.width()
                    case "stretch":
                        self.dims.content.width = parent.dims.content.width - self.dims.horizontal_edge_width()

        # calculate available width for children, minus how much they use,
        # then divide that between them based on the content justification
        # We are in the parent, justifying the children!

        available_width = self.dims.content.width
        available_height = self.dims.content.height

        relative_children = [child for child in self.children if child.element.style.layout.position == "relative"]
        relative_children_with_weights = [
            child for child in relative_children if child.element.style.layout.weight is not None
        ]
        num_relative_children = len(relative_children)
        num_gaps = max(sum(1 for child in self.children if child.element.style.layout.position == "relative") - 1, 0)
        total_gap = num_gaps * layout.gap_children

        # subtract off fixed-width/height children from what's available to flex
        for child in relative_children:
            if child.element.style.layout.weight is None:
                if layout.direction == "row":
                    available_width -= child.dims.width()
                elif layout.direction == "column":
                    available_height -= child.dims.height()

        # when does flex element width get set for space-* justify?
        # space-* justify assumes children have fixed widths! it distributes the leftover space
        # but what if you DO put flex children in there? who sets their widths?
        # TODO: note that if a child element has a weight, a fixed width/height would be overriden here by the parent
        if relative_children_with_weights and layout.justify_children not in (
            "space-between",
            "space-around",
            "space-evenly",
        ):
            weights: tuple[int] = tuple(child.element.style.layout.weight for child in relative_children_with_weights)  # type: ignore[assignment]
            if layout.direction == "row":
                available_width -= total_gap

                for child, flex_portion in zip(
                    relative_children_with_weights, partition_int(total=available_width, weights=weights)
                ):
                    child.dims.content.width = max(flex_portion - child.dims.horizontal_edge_width(), 0)

                available_width = 0

            elif layout.direction == "column":
                available_height -= total_gap

                for child, flex_portion in zip(
                    relative_children_with_weights, partition_int(total=available_height, weights=weights)
                ):
                    child.dims.content.height = max(flex_portion - child.dims.vertical_edge_width(), 0)

                available_height = 0

        elif layout.justify_children in ("space-between", "space-around", "space-evenly"):
            for child in relative_children:
                available_width -= child.dims.width()
                available_height -= child.dims.height()

        # at this point we know how wide each child is, so we can do text wrapping and set heights
        for child in relative_children:
            if child.element.type == "text" and child.element.style.typography.wrap != "none":
                h = len(
                    wrap_cells(
                        cells=child.element.cells,
                        wrap=child.element.style.typography.wrap,
                        width=child.dims.content.width,
                    )
                )
                child.dims.content.height = max(min(h, available_height - child.dims.vertical_edge_width()), 0)

        # determine positions

        # shift nominal content box position set by parent by own margin, border, and padding
        self.dims.content.x += self.dims.margin.left + self.dims.border.left + self.dims.padding.left
        self.dims.content.y += self.dims.margin.top + self.dims.border.top + self.dims.padding.top

        # these x and y persist between children and set the top left corner of their margin box
        # start in top left corner of our own context box
        x = self.dims.content.x
        y = self.dims.content.y

        # TODO: available width can be negative here, but that doesn't make sense
        # seems to happen when this element isn't stretch but it has fixed-width children

        # justification (main axis placement)
        # TODO: need to subtract total_gap here, even though it was handled above?
        if layout.direction == "row":
            if layout.justify_children == "center":
                x += (available_width - total_gap) // 2
            elif layout.justify_children == "end":
                x += available_width - total_gap
        elif layout.direction == "column":
            if layout.justify_children == "center":
                y += (available_height - total_gap) // 2
            elif layout.justify_children == "end":
                y += available_height - total_gap

        if layout.justify_children == "space-between":
            if layout.direction == "row":
                for child, gap in zip(
                    relative_children,
                    partition_int(available_width, (1,) * (num_relative_children - 1) + (0,)),
                ):
                    child.dims.content.x = x
                    child.dims.content.y = y
                    x += child.dims.width() + gap
            elif layout.direction == "column":
                for child, gap in zip(
                    relative_children,
                    partition_int(available_height, (1,) * (num_relative_children - 1) + (0,)),
                ):
                    child.dims.content.x = x
                    child.dims.content.y = y
                    y += child.dims.height() + gap

        elif layout.justify_children == "space-around":
            if layout.direction == "row":
                for child, gap in zip(
                    relative_children,
                    partition_int(available_width, (1,) * num_relative_children),
                ):
                    child.dims.content.x = x + halve_integer(gap)[0]
                    child.dims.content.y = y
                    x += child.dims.width() + gap
            elif layout.direction == "column":
                for child, gap in zip(
                    relative_children,
                    partition_int(available_height, (1,) * num_relative_children),
                ):
                    child.dims.content.x = x
                    child.dims.content.y = y + halve_integer(gap)[0]
                    y += child.dims.height() + gap

        elif layout.justify_children == "space-evenly":
            if layout.direction == "row":
                for child, gap in zip(
                    relative_children,
                    partition_int(available_width, (1,) * (num_relative_children + 1)),
                ):
                    child.dims.content.x = x + gap
                    child.dims.content.y = y
                    x += child.dims.width() + gap
            elif layout.direction == "column":
                for child, gap in zip(
                    relative_children,
                    partition_int(available_height, (1,) * (num_relative_children + 1)),
                ):
                    child.dims.content.x = x
                    child.dims.content.y = y + gap
                    y += child.dims.height() + gap

        else:
            for child in relative_children:
                child.dims.content.x = x
                child.dims.content.y = y
                if layout.direction == "row":
                    x += child.dims.width() + layout.gap_children
                elif layout.direction == "column":
                    y += child.dims.height() + layout.gap_children

        # alignment (cross-axis placement)
        # content width/height of self, but full width/height of children
        for child in relative_children:
            # Skip children that will align themselves
            if child.element.style.layout.align_self != "none":
                continue

            if layout.direction == "row":
                if layout.align_children == "center":
                    # TODO: these floordivs aren't great
                    child.dims.content.y = self.dims.content.y + ((self.dims.content.height - child.dims.height()) // 2)
                elif layout.align_children == "end":
                    child.dims.content.y = self.dims.content.y + self.dims.content.height - child.dims.height()
                elif layout.align_children == "stretch" and child.element.style.span.height == "auto":
                    child.dims.content.height = self.dims.content.height - child.dims.vertical_edge_width()
            elif layout.direction == "column":
                if layout.align_children == "center":
                    # TODO: these floordivs aren't great
                    child.dims.content.x = self.dims.content.x + ((self.dims.content.width - child.dims.width()) // 2)
                elif layout.align_children == "end":
                    child.dims.content.x = self.dims.content.y + self.dims.content.width - child.dims.width()
                elif layout.align_children == "stretch" and child.element.style.span.width == "auto":
                    child.dims.content.width = self.dims.content.width - child.dims.horizontal_edge_width()


def build_layout_tree(element: AnyElement, parent: LayoutBox | None = None) -> LayoutBox:
    if element.style.hidden:
        raise Exception("Root element cannot have layout='hidden'")

    box = LayoutBox(element=element, parent=parent)
    for child in element.children:
        if isinstance(child, Component):
            raise Exception("Layout tree must be built from concrete Elements, not Components")

        if not child.style.hidden == "hidden":
            box.children.append(build_layout_tree(element=child, parent=box))

    return box



from time import monotonic_ns
from typing import Union

from pydantic import Field

from counterweight.geometry import Position
from counterweight.types import FrozenForbidExtras


class _Event(FrozenForbidExtras):
    timestamp_ns: int = Field(default_factory=monotonic_ns)


class TerminalResized(_Event):
    pass


class KeyPressed(_Event):
    key: str


class MouseMoved(_Event):
    position: Position


class MouseDown(_Event):
    position: Position
    button: int


class MouseUp(_Event):
    position: Position


class StateSet(_Event):
    pass


class Dummy(_Event):
    pass


AnyEvent = Union[
    TerminalResized,
    KeyPressed,
    MouseMoved,
    MouseDown,
    MouseUp,
    StateSet,
    Dummy,
]



from __future__ import annotations

from collections.abc import Generator, Mapping
from enum import Enum
from string import printable

from parsy import Parser, any_char, char_from, decimal_digit, generate, string
from structlog import get_logger

from counterweight.events import AnyEvent, KeyPressed, MouseDown, MouseMoved, MouseUp
from counterweight.geometry import Position

logger = get_logger()


class Key(str, Enum):
    """
    Derived from https://github.com/Textualize/textual/blob/c966243b059f0352e2a23b9695776838195364a3/src/textual/keys.py
    """

    Escape = "escape"  # Also Control-[
    ShiftEscape = "shift+escape"
    Return = "return"

    ControlAt = "ctrl+@"  # Also Control-Space.

    ControlA = "ctrl+a"
    ControlB = "ctrl+b"
    ControlC = "ctrl+c"
    ControlD = "ctrl+d"
    ControlE = "ctrl+e"
    ControlF = "ctrl+f"
    ControlG = "ctrl+g"
    ControlH = "ctrl+h"
    ControlI = "ctrl+i"  # Tab
    ControlJ = "ctrl+j"  # Newline
    ControlK = "ctrl+k"
    ControlL = "ctrl+l"
    ControlM = "ctrl+m"  # Carriage return
    ControlN = "ctrl+n"
    ControlO = "ctrl+o"
    ControlP = "ctrl+p"
    ControlQ = "ctrl+q"
    ControlR = "ctrl+r"
    ControlS = "ctrl+s"
    ControlT = "ctrl+t"
    ControlU = "ctrl+u"
    ControlV = "ctrl+v"
    ControlW = "ctrl+w"
    ControlX = "ctrl+x"
    ControlY = "ctrl+y"
    ControlZ = "ctrl+z"

    Control1 = "ctrl+1"
    Control2 = "ctrl+2"
    Control3 = "ctrl+3"
    Control4 = "ctrl+4"
    Control5 = "ctrl+5"
    Control6 = "ctrl+6"
    Control7 = "ctrl+7"
    Control8 = "ctrl+8"
    Control9 = "ctrl+9"
    Control0 = "ctrl+0"

    ControlShift1 = "ctrl+shift+1"
    ControlShift2 = "ctrl+shift+2"
    ControlShift3 = "ctrl+shift+3"
    ControlShift4 = "ctrl+shift+4"
    ControlShift5 = "ctrl+shift+5"
    ControlShift6 = "ctrl+shift+6"
    ControlShift7 = "ctrl+shift+7"
    ControlShift8 = "ctrl+shift+8"
    ControlShift9 = "ctrl+shift+9"
    ControlShift0 = "ctrl+shift+0"

    ControlBackslash = "ctrl+backslash"
    ControlSquareClose = "ctrl+right_square_bracket"
    ControlCircumflex = "ctrl+circumflex_accent"
    ControlUnderscore = "ctrl+underscore"

    Left = "left"
    Right = "right"
    Up = "up"
    Down = "down"
    Home = "home"
    End = "end"
    Insert = "insert"
    Delete = "delete"
    PageUp = "pageup"
    PageDown = "pagedown"

    ControlLeft = "ctrl+left"
    ControlRight = "ctrl+right"
    ControlUp = "ctrl+up"
    ControlDown = "ctrl+down"
    ControlHome = "ctrl+home"
    ControlEnd = "ctrl+end"
    ControlInsert = "ctrl+insert"
    ControlDelete = "ctrl+delete"
    ControlPageUp = "ctrl+pageup"
    ControlPageDown = "ctrl+pagedown"

    ShiftLeft = "shift+left"
    ShiftRight = "shift+right"
    ShiftUp = "shift+up"
    ShiftDown = "shift+down"
    ShiftHome = "shift+home"
    ShiftEnd = "shift+end"
    ShiftInsert = "shift+insert"
    ShiftDelete = "shift+delete"
    ShiftPageUp = "shift+pageup"
    ShiftPageDown = "shift+pagedown"

    ControlShiftLeft = "ctrl+shift+left"
    ControlShiftRight = "ctrl+shift+right"
    ControlShiftUp = "ctrl+shift+up"
    ControlShiftDown = "ctrl+shift+down"
    ControlShiftHome = "ctrl+shift+home"
    ControlShiftEnd = "ctrl+shift+end"
    ControlShiftInsert = "ctrl+shift+insert"
    ControlShiftDelete = "ctrl+shift+delete"
    ControlShiftPageUp = "ctrl+shift+pageup"
    ControlShiftPageDown = "ctrl+shift+pagedown"

    AltDelete = "alt+delete"

    BackTab = "shift+tab"  # shift + tab

    F1 = "f1"
    F2 = "f2"
    F3 = "f3"
    F4 = "f4"
    F5 = "f5"
    F6 = "f6"
    F7 = "f7"
    F8 = "f8"
    F9 = "f9"
    F10 = "f10"
    F11 = "f11"
    F12 = "f12"
    F13 = "f13"
    F14 = "f14"
    F15 = "f15"
    F16 = "f16"
    F17 = "f17"
    F18 = "f18"
    F19 = "f19"
    F20 = "f20"
    F21 = "f21"
    F22 = "f22"
    F23 = "f23"
    F24 = "f24"

    ControlF1 = "ctrl+f1"
    ControlF2 = "ctrl+f2"
    ControlF3 = "ctrl+f3"
    ControlF4 = "ctrl+f4"
    ControlF5 = "ctrl+f5"
    ControlF6 = "ctrl+f6"
    ControlF7 = "ctrl+f7"
    ControlF8 = "ctrl+f8"
    ControlF9 = "ctrl+f9"
    ControlF10 = "ctrl+f10"
    ControlF11 = "ctrl+f11"
    ControlF12 = "ctrl+f12"
    ControlF13 = "ctrl+f13"
    ControlF14 = "ctrl+f14"
    ControlF15 = "ctrl+f15"
    ControlF16 = "ctrl+f16"
    ControlF17 = "ctrl+f17"
    ControlF18 = "ctrl+f18"
    ControlF19 = "ctrl+f19"
    ControlF20 = "ctrl+f20"
    ControlF21 = "ctrl+f21"
    ControlF22 = "ctrl+f22"
    ControlF23 = "ctrl+f23"
    ControlF24 = "ctrl+f24"

    # Matches any key.
    Any = "<any>"

    # Special.
    ScrollUp = "<scroll-up>"
    ScrollDown = "<scroll-down>"

    # For internal use: key which is ignored.
    # (The key binding for this key should not do anything.)
    Ignore = "<ignore>"

    # Some 'Key' aliases (for backward-compatibility).
    ControlSpace = "ctrl-at"
    Tab = "tab"
    Space = "space"
    Enter = "enter"
    Backspace = "backspace"

    def __repr__(self) -> str:
        return str(self)


SINGLE_CHAR_TRANSFORMS: Mapping[str, Key] = {
    "\x1b": Key.Escape,
    "\t": Key.Tab,
    "\n": Key.Enter,
    " ": Key.Space,
    "\x00": Key.ControlSpace,
    "\x01": Key.ControlA,
    "\x02": Key.ControlB,
    "\x03": Key.ControlC,
    "\x04": Key.ControlD,
    "\x05": Key.ControlE,
    "\x06": Key.ControlF,
    "\x07": Key.ControlG,
    "\x08": Key.Backspace,
    "\x0B": Key.ControlK,
    "\x0C": Key.ControlL,
    "\x0E": Key.ControlN,
    "\x0F": Key.ControlO,
    "\x10": Key.ControlP,
    "\x11": Key.ControlQ,
    "\x12": Key.ControlR,
    "\x13": Key.ControlS,
    "\x14": Key.ControlT,
    "\x15": Key.ControlU,
    "\x16": Key.ControlV,
    "\x17": Key.ControlW,
    "\x18": Key.ControlX,
    "\x19": Key.ControlY,
    "\x1A": Key.ControlZ,
    "\x7f": Key.Backspace,
}

CHARS = "".join((*printable, *SINGLE_CHAR_TRANSFORMS.keys()))


@generate
def single_char() -> Generator[Parser, str, AnyEvent]:
    c = yield char_from(CHARS)

    return KeyPressed(key=SINGLE_CHAR_TRANSFORMS.get(c, c))


@generate
def escape_sequence() -> Generator[Parser, AnyEvent, AnyEvent]:
    keys = yield string("\x1b") >> (
        f1to4 | (string("[") >> (mouse_position | mouse_button | shift_tab | two_params | zero_or_one_params))
    )

    return keys


F1TO4 = {
    "P": Key.F1,
    "Q": Key.F2,
    "R": Key.F3,
    "S": Key.F4,
}


@generate
def f1to4() -> Generator[Parser, str, AnyEvent]:
    yield string("O")
    final = yield char_from("PQRS")

    return KeyPressed(key=F1TO4[final])


@generate
def shift_tab() -> Generator[Parser, str, AnyEvent]:
    yield string("Z")

    return KeyPressed(key=Key.BackTab)


@generate
def mouse_position() -> Generator[Parser, str, AnyEvent]:
    # https://www.xfree86.org/current/ctlseqs.html
    yield string("MC")

    x_char = yield any_char
    y_char = yield any_char

    x = ord(x_char) - 33
    y = ord(y_char) - 33

    return MouseMoved(position=Position(x=x, y=y))


@generate
def mouse_button() -> Generator[Parser, str, AnyEvent]:
    # https://www.xfree86.org/current/ctlseqs.html
    yield string("M")

    buttons_char = yield any_char

    buttons = ord(buttons_char) - 32

    x_char = yield any_char
    y_char = yield any_char

    x = ord(x_char) - 33
    y = ord(y_char) - 33
    p = Position(x=x, y=y)

    # TODO: handle upper bits of buttons
    if buttons & 3 == 3:
        return MouseUp(position=p)
    elif buttons & 2 == 2:
        return MouseDown(position=p, button=3)
    elif buttons & 1 == 1:
        return MouseDown(position=p, button=2)
    else:  # low bits are 00, so this is mouse down with button 1
        return MouseDown(position=p, button=1)


CSI_LOOKUP: Mapping[tuple[str, ...], str] = {
    ("", "A"): Key.Up,
    ("", "B"): Key.Down,
    ("", "C"): Key.Right,
    ("", "D"): Key.Left,
    ("", "F"): Key.End,
    ("2", "~"): Key.Insert,
    ("3", "~"): Key.Delete,
    ("11", "~"): Key.F1,
    ("12", "~"): Key.F2,
    ("13", "~"): Key.F3,
    ("14", "~"): Key.F4,
    ("15", "~"): Key.F5,
    # skip 16
    ("17", "~"): Key.F6,
    ("18", "~"): Key.F7,
    ("19", "~"): Key.F8,
    ("20", "~"): Key.F9,
    ("21", "~"): Key.F10,
    # skip 22
    ("23", "~"): Key.F11,
    ("24", "~"): Key.F12,
    ("25", "~"): Key.F13,
    ("26", "~"): Key.F14,
    ("28", "~"): Key.F15,
    ("29", "~"): Key.F16,
    # skip 30
    ("31", "~"): Key.F17,
    ("32", "~"): Key.F18,
    ("33", "~"): Key.F19,
    ("34", "~"): Key.F20,
    ("1", "2", "A"): Key.ShiftUp,
    ("1", "2", "B"): Key.ShiftDown,
    ("1", "2", "C"): Key.ShiftRight,
    ("1", "2", "D"): Key.ShiftLeft,
    ("1", "5", "A"): Key.ControlUp,
    ("1", "5", "B"): Key.ControlDown,
    ("1", "5", "C"): Key.ControlRight,
    ("1", "5", "D"): Key.ControlLeft,
    ("1", "6", "A"): Key.ControlShiftUp,
    ("1", "6", "B"): Key.ControlShiftDown,
    ("1", "6", "C"): Key.ControlShiftRight,
    ("1", "6", "D"): Key.ControlShiftLeft,
    ("3", "3", "~"): Key.AltDelete,
    ("3", "5", "~"): Key.ControlDelete,
    ("3", "6", "~"): Key.ControlShiftInsert,
}

FINAL_CHARS = "".join(sorted(set(key[-1] for key in CSI_LOOKUP)))


@generate
def two_params() -> Generator[Parser, str, AnyEvent]:
    p1 = yield decimal_digit.many().concat()
    yield string(";")
    p2 = yield decimal_digit.many().concat()
    e = yield char_from(FINAL_CHARS)

    return KeyPressed(key=CSI_LOOKUP[(p1, p2, e)])


@generate
def zero_or_one_params() -> Generator[Parser, str, AnyEvent]:
    # zero params => ""
    p1 = yield decimal_digit.many().concat()

    e = yield char_from(FINAL_CHARS)

    return KeyPressed(key=CSI_LOOKUP[(p1, e)])


@generate
def vt_inputs() -> Generator[Parser, list[AnyEvent], list[AnyEvent]]:
    commands = yield (escape_sequence | single_char).many()
    return commands



from __future__ import annotations

import os
import selectors
import termios
from collections.abc import Callable
from copy import deepcopy
from selectors import DefaultSelector
from time import perf_counter_ns
from typing import TextIO

from parsy import ParseError
from structlog import get_logger

from counterweight.events import AnyEvent
from counterweight.keys import vt_inputs

logger = get_logger()


def read_keys(stream: TextIO, put_event: Callable[[AnyEvent], None]) -> None:
    """
    Based on https://github.com/Textualize/textual/blob/bb9cc6281aa717054c8133ce4a2eac5ad082c574/src/textual/drivers/linux_driver.py#L236
    """
    selector = DefaultSelector()
    selector.register(stream, selectors.EVENT_READ)

    while True:
        selector.select()
        # We're just reading from one file,
        # so we can dispense with the ceremony of actually using the results of the select.

        start_parsing = perf_counter_ns()
        b = os.read(stream.fileno(), 1_000)
        bytes = list(b)
        buffer = b.decode("utf-8")
        try:
            inputs = vt_inputs.parse(buffer)

            for i in inputs:
                put_event(i)

            logger.debug(
                "Parsed user input",
                inputs=inputs,
                buffer=repr(buffer),
                bytes=bytes,
                len_buffer=len(buffer),
                elapsed_ns=f"{perf_counter_ns() - start_parsing:_}",
            )
        except (ParseError, KeyError) as e:
            logger.error(
                "Failed to parse input",
                error=str(e),
                buffer=repr(buffer),
                len_buffer=len(buffer),
                elapsed_ns=f"{perf_counter_ns() - start_parsing:_}",
            )


LFLAG = 3
CC = 6

TCGetAttr = list[int | list[int | bytes]]


def start_input_control(stream: TextIO) -> TCGetAttr:
    original = termios.tcgetattr(stream)

    modified = deepcopy(original)

    modified[LFLAG] = original[LFLAG] & ~(termios.ECHO | termios.ICANON)
    modified[CC][termios.VMIN] = 1
    modified[CC][termios.VTIME] = 0

    termios.tcsetattr(stream.fileno(), termios.TCSADRAIN, modified)

    return original


def stop_input_control(stream: TextIO, original: TCGetAttr) -> None:
    termios.tcsetattr(stream.fileno(), termios.TCSADRAIN, original)



from __future__ import annotations

from functools import wraps
from typing import Callable, ParamSpec

from counterweight.elements import AnyElement
from counterweight.types import FrozenForbidExtras

P = ParamSpec("P")

Key = str | int | None


def component(func: Callable[P, AnyElement]) -> Callable[P, Component]:
    """
    A decorator that marks a function as a component.
    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Component:
        return Component(func=func, args=args, kwargs=kwargs)

    return wrapper


class Component(FrozenForbidExtras):
    """
    The result of calling a component function.
    These should not be instantiated directly;
    instead, use the `@component` decorator on a function
    and call it normally.
    """

    func: Callable[..., AnyElement]
    args: tuple[object, ...]
    kwargs: dict[str, object]
    key: Key = None

    def with_key(self, key: Key) -> Component:
        return self.model_copy(update={"key": key})


Component.model_rebuild()



from __future__ import annotations

from typing import Callable, Iterator, Literal, Sequence, Union

from pydantic import Field

from counterweight.cell_paint import CellPaint
from counterweight.controls import AnyControl
from counterweight.events import KeyPressed, MouseDown, MouseUp
from counterweight.styles import CellStyle, Style
from counterweight.types import FrozenForbidExtras


class Div(FrozenForbidExtras):
    type: Literal["div"] = "div"
    style: Style = Field(default=Style())
    children: Sequence[Component | AnyElement] = Field(default=())
    on_hover: Style = Field(default=Style())
    on_key: Callable[[KeyPressed], AnyControl | None] | None = None
    on_mouse_down: Callable[[MouseDown], AnyControl | None] | None = None
    on_mouse_up: Callable[[MouseUp], AnyControl | None] | None = None


class Chunk(FrozenForbidExtras):
    content: str
    style: CellStyle = Field(default_factory=CellStyle)

    @property
    def cells(self) -> Iterator[CellPaint]:
        yield from (CellPaint(char=char, style=self.style) for char in self.content)

    @classmethod
    def space(cls) -> Chunk:
        return Chunk(content=" ")

    @classmethod
    def newline(cls) -> Chunk:
        return Chunk(content="\n")


class Text(FrozenForbidExtras):
    type: Literal["text"] = "text"
    content: str | Sequence[Chunk | CellPaint]
    style: Style = Field(default=Style())
    on_hover: Style | None = Field(default=None)
    on_key: Callable[[KeyPressed], AnyControl | None] | None = None
    on_mouse_down: Callable[[MouseDown], AnyControl | None] | None = None
    on_mouse_up: Callable[[MouseUp], AnyControl | None] | None = None

    @property
    def children(self) -> Sequence[Component | AnyElement]:
        return ()

    @property
    def cells(self) -> Iterator[CellPaint]:
        if isinstance(self.content, str):
            yield from (CellPaint(char=char, style=self.style.typography.style) for char in self.content)
        else:
            for c in self.content:
                yield from c.cells


AnyElement = Union[
    Div,
    Text,
]

from counterweight.components import Component  # noqa: E402, deferred to avoid circular import

Div.model_rebuild()
Text.model_rebuild()



from __future__ import annotations

from asyncio import Queue
from contextvars import ContextVar
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from counterweight.events import AnyEvent
    from counterweight.hooks.impls import Hooks

current_event_queue: ContextVar[Queue[AnyEvent]] = ContextVar("current_event_queue")
current_hook_idx: ContextVar[int] = ContextVar("current_hook_idx")
current_hook_state: ContextVar[Hooks] = ContextVar("current_hook_state")



from __future__ import annotations

import sys
from asyncio import Queue, get_running_loop, run
from textwrap import dedent
from threading import Thread

from typer import Option, Typer

from counterweight._context_vars import current_event_queue
from counterweight.constants import PACKAGE_NAME, __version__
from counterweight.events import AnyEvent
from counterweight.input import read_keys, start_input_control, stop_input_control
from counterweight.logging import tail_devlog
from counterweight.output import start_mouse_reporting, stop_mouse_reporting

cli = Typer(
    name=PACKAGE_NAME,
    no_args_is_help=True,
    rich_markup_mode="rich",
    help=dedent(
        """\
        """
    ),
)


@cli.command()
def version() -> None:
    """
    Display version information.
    """
    print(__version__)


@cli.command()
def devlog() -> None:
    """Tail the developer log file."""
    tail_devlog()


@cli.command()
def check_input(mouse: bool = Option(default=False, help="Also capture mouse inputs and show mouse events.")) -> None:
    """
    Enter the same input-capture state used during application mode,
    and show the results of reading input (e.g., key events).
    """

    run(_check_input(mouse=mouse))


async def _check_input(mouse: bool) -> None:
    event_queue: Queue[AnyEvent] = Queue()
    current_event_queue.set(event_queue)

    loop = get_running_loop()

    def put_event(event: AnyEvent) -> None:
        loop.call_soon_threadsafe(event_queue.put_nowait, event)

    input_stream = sys.stdin
    output_stream = sys.stdout

    key_thread = Thread(target=read_keys, args=(input_stream, put_event), daemon=True)
    key_thread.start()

    original = start_input_control(stream=input_stream)
    if mouse:
        start_mouse_reporting(stream=output_stream)
    try:
        while True:
            print("Waiting for input...")
            key = await event_queue.get()
            print(f"Event: {key!r}")
    except KeyboardInterrupt:
        print("Exiting...")
    finally:
        stop_input_control(stream=input_stream, original=original)
        if mouse:
            stop_mouse_reporting(stream=output_stream)



from __future__ import annotations

from collections.abc import Iterable
from typing import Iterator, Literal

from pydantic import Field

from counterweight.styles.styles import CellStyle
from counterweight.types import FrozenForbidExtras


class CellPaint(FrozenForbidExtras):
    char: str = Field(default=..., min_length=1, max_length=1)
    style: CellStyle = Field(default_factory=CellStyle)

    @property
    def cells(self) -> Iterator[CellPaint]:
        yield self


def wrap_cells(
    cells: Iterable[CellPaint],
    wrap: Literal["none", "paragraphs"],
    width: int,
) -> list[list[CellPaint]]:
    if width <= 0:
        return []

    if wrap == "none":
        lines: list[list[CellPaint]] = []
        current_line: list[CellPaint] = []
        for cell in cells:
            if cell.char == "\n":
                lines.append(current_line)
                current_line = []
            else:
                current_line.append(cell)
        lines.append(current_line)
        return lines

    raise NotImplementedError("non-none wrapping not yet implemented")

    # wrapper = TextWrapper(width=width)
    #
    # paragraphs = text.split("\n\n")  # double newline = paragraph break
    #
    # lines = []
    # for paragraph in paragraphs:
    #     lines.extend(wrapper.wrap(paragraph))
    #     lines.append("")  # empty line between paragraphs
    #
    # return lines[:-1]  # remove last empty line



from counterweight.hooks.hooks import use_effect, use_ref, use_state
from counterweight.hooks.types import Deps, Getter, Ref, Setter, Setup

__all__ = [
    "use_state",
    "use_ref",
    "use_effect",
    "Getter",
    "Setter",
    "Setup",
    "Deps",
    "Ref",
]



from __future__ import annotations

from collections import defaultdict
from textwrap import dedent
from typing import Literal, assert_never
from xml.etree.ElementTree import Element, ElementTree, SubElement

from structlog import get_logger

from counterweight._utils import halve_integer
from counterweight.cell_paint import CellPaint, wrap_cells
from counterweight.elements import AnyElement, Div, Text
from counterweight.geometry import Position
from counterweight.layout import BoxDimensions, Edge, LayoutBox, Rect
from counterweight.styles import Border
from counterweight.styles.styles import BorderEdge, CellStyle, Color, Margin, Padding

logger = get_logger()

Paint = dict[Position, CellPaint]


def paint_layout(layout: LayoutBox) -> Paint:
    painted = paint_element(layout.element, layout.dims)
    for child in layout.children:
        painted |= paint_layout(child)  # no Z-level support! need something like a chainmap
    return painted


def paint_element(element: AnyElement, dims: BoxDimensions) -> Paint:
    padding_rect, border_rect, margin_rect = dims.padding_border_margin_rects()
    m = paint_edge(element.style.margin, dims.margin, margin_rect)
    b = paint_border(element.style.border, border_rect) if element.style.border else {}
    t = paint_edge(element.style.padding, dims.padding, padding_rect)

    box = m | b | t

    match element:
        case Div():
            return box
        case Text() as e:
            return paint_text(e, dims.content) | box
        case _:
            raise NotImplementedError(f"Painting {element} is not implemented")


def justify_line(line: list[CellPaint], width: int, justify: Literal["left", "right", "center"]) -> list[CellPaint]:
    space = width - len(line)
    if space <= 0:
        return line
    elif justify == "left":
        return line + [CellPaint(char=" ")] * space
    elif justify == "right":
        return [CellPaint(char=" ")] * space + line
    elif justify == "center":
        left, right = halve_integer(space)
        return [CellPaint(char=" ")] * left + line + [CellPaint(char=" ")] * right
    else:
        assert_never(justify)


def paint_text(text: Text, rect: Rect) -> Paint:
    style = text.style.typography.style

    paint = {}
    lines = wrap_cells(
        cells=text.cells,
        wrap=text.style.typography.wrap,
        width=rect.width,
    )

    previous_cell_style = None

    for y, line in enumerate(lines[: rect.height], start=rect.y):
        justified_line = justify_line(line, rect.width, text.style.typography.justify)
        for x, cell in enumerate(justified_line[: rect.width], start=rect.x):
            pos = Position(x, y)
            cell_style = cell.style

            # Optimization: reuse the same merged style object if the cell style is the same object.
            # This helps when painting a lot of text with the same style.
            if cell_style is not previous_cell_style:
                merged_style = style | cell_style
                previous_cell_style = cell_style

            paint[pos] = CellPaint(
                char=cell.char,
                style=merged_style,  # merged_style will never be unassigned here, since we know previous_cell_style starts as None
            )

    return paint


def paint_edge(mp: Margin | Padding, edge: Edge, rect: Rect, char: str = " ") -> Paint:
    cell_paint = CellPaint(char=char, style=CellStyle(background=mp.color))

    chars = {}

    # top
    for y in range(rect.top, rect.top + edge.top):
        for x in rect.x_range():
            chars[Position(x, y)] = cell_paint

    # bottom
    for y in range(rect.bottom, rect.bottom - edge.bottom, -1):
        for x in rect.x_range():
            chars[Position(x, y)] = cell_paint

    # left
    for x in range(rect.left, rect.left + edge.left):
        for y in rect.y_range():
            chars[Position(x, y)] = cell_paint

    # right
    for x in range(rect.right, rect.right - edge.right, -1):
        for y in rect.y_range():
            chars[Position(x, y)] = cell_paint

    return chars


def paint_border(border: Border, rect: Rect) -> Paint:
    style = border.style

    bv = border.kind.value
    left = bv.left
    right = bv.right
    top = bv.top
    bottom = bv.bottom
    left_top = bv.left_top
    right_top = bv.right_top
    left_bottom = bv.left_bottom
    right_bottom = bv.right_bottom

    contract = border.contract
    chars = {}

    rect_left = rect.left
    rect_top = rect.top
    rect_right = rect.right
    rect_bottom = rect.bottom

    draw_left = BorderEdge.Left in border.edges
    draw_right = BorderEdge.Right in border.edges
    draw_top = BorderEdge.Top in border.edges
    draw_bottom = BorderEdge.Bottom in border.edges

    if contract:
        contract_top = contract if not draw_top else None
        contract_bottom = -contract if not draw_bottom else None
        contract_left = contract if not draw_left else None
        contract_right = -contract if not draw_right else None
    else:
        contract_top = contract_bottom = contract_left = contract_right = None

    v_slice = slice(contract_top, contract_bottom)
    h_slice = slice(contract_left, contract_right)

    if draw_left:
        left_paint = CellPaint(char=left, style=style)
        for p in rect.left_edge()[v_slice]:
            chars[p] = left_paint

    if draw_right:
        right_paint = CellPaint(char=right, style=style)
        for p in rect.right_edge()[v_slice]:
            chars[p] = right_paint

    if draw_top:
        top_paint = CellPaint(char=top, style=style)
        for p in rect.top_edge()[h_slice]:
            chars[p] = top_paint

    if draw_bottom:
        bottom_paint = CellPaint(char=bottom, style=style)
        for p in rect.bottom_edge()[h_slice]:
            chars[p] = bottom_paint

    if draw_top:
        if draw_left:
            chars[Position(x=rect_left, y=rect_top)] = CellPaint(char=left_top, style=style)
        if draw_right:
            chars[Position(x=rect_right, y=rect_top)] = CellPaint(char=right_top, style=style)

    if draw_bottom:
        if draw_left:
            chars[Position(x=rect_left, y=rect_bottom)] = CellPaint(char=left_bottom, style=style)
        if draw_right:
            chars[Position(x=rect_right, y=rect_bottom)] = CellPaint(char=right_bottom, style=style)

    return chars


def debug_paint(paint: dict[Position, CellPaint], rect: Rect) -> str:
    lines = []
    for y in rect.y_range():
        line = []
        for x in rect.x_range():
            line.append(paint.get(Position(x, y), CellPaint(char=" ")).char)

        lines.append(line)

    return "\n".join("".join(line) for line in lines)


def svg(paint: Paint) -> ElementTree:
    w, h = max(paint.keys())

    # Measurements start from the top-left corner of each cell, so the width/height need be 1 unit larger for the actual content
    w += 1
    h += 1

    x_mul = 0.55  # x coordinates get cut roughly in half because monospace cells are twice as tall as they are wide
    y_mul = 1.13  # seems to make border connect up just right

    fmt = "0.2f"
    unit = "em"

    root = Element(
        "svg",
        {
            "xmlns": "http://www.w3.org/2000/svg",
            "width": f"{w * x_mul:{fmt}}{unit}",
            "height": f"{h * y_mul:{fmt}}{unit}",
        },
    )

    style = SubElement(
        root,
        "style",
        {},
    )
    style.text = dedent(
        """\
        svg {
          dominant-baseline: hanging;
          font-size: 1rem;
        }
        """
    )

    background_root = SubElement(
        root,
        "g",
        {},
    )
    SubElement(  # default background color is black, so do a single black rectangle to cover the whole background as an optimization
        background_root,
        "rect",
        {
            "x": f"{0 * x_mul:{fmt}}{unit}",
            "y": f"{0 * y_mul:{fmt}}{unit}",
            "width": f"{w * x_mul:{fmt}}{unit}",
            "height": f"{h * y_mul:{fmt}}{unit}",
            "fill": Color.from_name("black").hex,
        },
    )

    text_root = SubElement(
        root,
        "text",
        {
            "fill": Color.from_name("white").hex,
            "font-family": "monospace",
        },
    )

    rows = defaultdict(list)
    for pos, cell in sorted(paint.items()):
        rows[pos.y].append((pos.x, cell))

    # Sort by y, then x, so that the SVG is written top to bottom, left to right
    for y, cells in sorted(rows.items()):
        row_root = SubElement(
            text_root,
            "tspan",
            {
                "y": f"{y * y_mul:{fmt}}{unit}",
            },
        )
        for x, cell in cells:
            # black is the default background color, so don't write it (optimization)
            if cell.style.background != Color.from_name("black"):
                SubElement(
                    background_root,
                    "rect",
                    {
                        "x": f"{x * x_mul:{fmt}}{unit}",
                        "width": f"{1.05 * x_mul:{fmt}}{unit}",  # go over the edge a bit on the right to cover gaps
                        "height": f"{1 * y_mul:{fmt}}{unit}",
                        "fill": cell.style.background.hex,
                    },
                )

            if cell.char == " ":  # optimization: don't write spaces
                continue

            ts = SubElement(
                row_root,
                "tspan",
                {
                    "x": f"{x * x_mul:{fmt}}{unit}",
                },
            )
            if cell.style.foreground != Color.from_name("white"):  # optimization: don't write white, it's the default
                ts.attrib["fill"] = cell.style.foreground.hex
            ts.text = cell.char

    return ElementTree(element=root)



from __future__ import annotations

from functools import lru_cache

from counterweight.cell_paint import CellPaint
from counterweight.geometry import Position
from counterweight.paint import Paint
from counterweight.styles.styles import JoinedBorderKind, JoinedBorderParts


@lru_cache(maxsize=2**12)
def get_replacement_char(
    parts: JoinedBorderParts,
    center: str,
    left: str | None,
    right: str | None,
    above: str | None,
    below: str | None,
) -> str | None:
    return parts.select(
        top=center in parts.connects_top or above in parts.connects_bottom,
        bottom=center in parts.connects_bottom or below in parts.connects_top,
        left=center in parts.connects_left or left in parts.connects_right,
        right=center in parts.connects_right or right in parts.connects_left,
    )


@lru_cache(maxsize=2**12)
def dither(position: Position) -> tuple[Position, Position, Position, Position]:
    return (
        Position(position.x - 1, position.y),  # left
        Position(position.x + 1, position.y),  # right
        Position(position.x, position.y - 1),  # above
        Position(position.x, position.y + 1),  # below
    )


JOINED_BORDER_KINDS = tuple(k.value for k in JoinedBorderKind)


def heal_borders(paint: Paint) -> Paint:
    overlay: Paint = {}
    for p, cell_paint in paint.items():
        char = cell_paint.char
        for kind in JOINED_BORDER_KINDS:
            if char not in kind:  # the center character must be a joined border part
                continue

            left, right, above, below = map(paint.get, dither(p))

            # TODO: cell styles must match too (i.e., colors)

            if replaced_char := get_replacement_char(
                kind,
                center=char,
                left=left.char if left else None,
                right=right.char if right else None,
                above=above.char if above else None,
                below=below.char if below else None,
            ):
                overlay[p] = CellPaint(char=replaced_char, style=cell_paint.style)

    return overlay
